from sys import modules
from typing import Dict, Any


class Methods:
    CLASSES = {}

    @staticmethod
    def get_class_attrs(target_cls: type) -> Dict[str, Any]:
        """
        Returns a dict of {name: value}, containing all attributes bound to a class
        """

        return {k: v for k, v in vars(target_cls).items() if not k.startswith("__")}

    @staticmethod
    def get_obj_by_str_name(object_name: str):
        """
        Returns the value stored under the provided name, in the global namespace.
        Supports dot notation to access objects stored under module or class attributes
        """

        nodes = object_name.split(".")

        obj = getattr(modules["__main__"], nodes.pop(0))
        for node in nodes:
            obj = getattr(obj, node)

        return obj



from typing import Callable

from ..screens import Screen
from ..room import Room


class RemoveObserver(Exception):
    pass


class ObserverHandler:
    REGISTERED = {
        "on_change_screen": set(),
        "on_change_room": set(),
    }

    @staticmethod
    def add(event_key: str, observer: Callable) -> None:
        ObserverHandler.REGISTERED[event_key].add(observer)

    @staticmethod
    def remove(event_key: str, observer: Callable) -> None:
        ObserverHandler.REGISTERED[event_key].remove(observer)

    @staticmethod
    def on_change_screen(old_screen: Screen, new_screen: Screen) -> None:
        for observer in ObserverHandler.REGISTERED["on_change_screen"]:
            try:
                observer(old_screen, new_screen)
            except RemoveObserver:
                ObserverHandler.remove("on_change_screen", observer)

    @staticmethod
    def on_change_room(old_room: Room, new_room: Room) -> None:
        for observer in ObserverHandler.REGISTERED["on_change_room"]:
            try:
                observer(old_room, new_room)
            except RemoveObserver:
                ObserverHandler.remove("on_change_screen", observer)



class Constants:
    RESOURCE_FOLDER_PATH = "res"

    AUDIO_FREQUENCY = 88200

    COLOURS = {
        "dev": (0, 255, 0)
    }


class ErrorMessages:
    @staticmethod
    def stat_locked():
        raise PermissionError("cannot modify this object while .is_locked is set to a truthy value")



from typing import Iterable
from abc import ABC
from weakref import ref

from ..tagged import Tagged
from .constants import Constants


class Hitbox(Tagged, ABC):
    def __init__(self, parent: "Renderable.with_extensions(Hitboxed)", tags: Iterable[str]):
        super().__init__(tags)

        # Weakref so that it does not prevent parent object being garbage collected
        self._parent_renderable = ref(parent)

    @property
    def hitbox_handler(self) -> "HitboxHandler":
        """
        Shortcut property which accesses the current game screen, and then the current hitbox handler through that
        """

        return self.parent_renderable.game.screen.hitbox_handler

    @property
    def parent_renderable(self) -> "Renderable.with_extensions(Hitboxed)":
        return self._parent_renderable()

    def is_collision(self, other: "Hitbox", check_by_parent: bool = True) -> bool:
        """
        Checks for a collision between this hitbox and the provided other hitbox.
        If this collision has already been checked this tick, returns False.

        If check_by_parent is True, hitbox objects that have the same parent Renderable
        are considered to be the same hitbox for this purpose
        """

        if check_by_parent:
            collision_key = frozenset((self.parent_renderable, other.parent_renderable))
        else:
            collision_key = frozenset((self, other))

        if collision_key in self.hitbox_handler.checked_collisions:
            return False  # This collision check has already been carried out previously this tick

        self.hitbox_handler.checked_collisions.add(collision_key)
        return self._is_collision(other)

    def _is_collision(self, other: "Hitbox") -> bool:
        """
        Must be overridden.
        Should complete (or further delegate, if necessary) the actual check for a collision
        between this hitbox object and the other provided hitbox object
        """

        raise NotImplementedError

    def _is_valid_tag(self, tag: str) -> bool:
        return tag in Constants.HITBOX_TAGS
